////////////////////////////////////////////////////////////////////////////
//
//
//    Project     : VideoNet version 1.1.
//    Description : Peer to Peer Video Conferencing over the LAN.
//    Author	  : Nagareshwar Y Talekar ( nsry2002@yahoo.co.in)
//    Date  	  : 15-6-2004.
//
//
//    File description : 
//    Name    : convert.cpp
//    Details : Conversion routine from RGB24 to YUV420 & YUV420 to RGB24.
//
/////////////////////////////////////////////////////////////////////////////


#include "convert.h"

// Conversion from RGB to YUV420
int RGB2YUV_YR[256], RGB2YUV_YG[256], RGB2YUV_YB[256];
int RGB2YUV_UR[256], RGB2YUV_UG[256], RGB2YUV_UBVR[256];
int RGB2YUV_VG[256], RGB2YUV_VB[256];

// Conversion from YUV420 to RGB24
static long int crv_tab[256];
static long int cbu_tab[256];
static long int cgu_tab[256];
static long int cgv_tab[256];
static long int tab_76309[256];
static unsigned char clp[1024]; 		//for clip in CCIR601

#define Limite8(x) \
	((int)((x>=0) ? ((x<=0xFF) ? x : 0xFF) : 0))
#define Limite1024(x) \
	((int)((x>0) ? ((x<1024) ? x : 1024) : 0))

//
// Table used for RGB to YUV420 conversion
//
void InitLookupTable()
{
	int i;

	for (i = 0; i < 256; i++)
		RGB2YUV_YR[i] = (float)65.481 * (i << 8);
	for (i = 0; i < 256; i++)
		RGB2YUV_YG[i] = (float)128.553 * (i << 8);
	for (i = 0; i < 256; i++)
		RGB2YUV_YB[i] = (float)24.966 * (i << 8);
	for (i = 0; i < 256; i++)
		RGB2YUV_UR[i] = (float)37.797 * (i << 8);
	for (i = 0; i < 256; i++)
		RGB2YUV_UG[i] = (float)74.203 * (i << 8);
	for (i = 0; i < 256; i++)
		RGB2YUV_VG[i] = (float)93.786 * (i << 8);
	for (i = 0; i < 256; i++)
		RGB2YUV_VB[i] = (float)18.214 * (i << 8);
	for (i = 0; i < 256; i++)
		RGB2YUV_UBVR[i] = (float)112 * (i << 8);
}


//
//  Convert from  RGB24 to YUV420
//
int ConvertRGB2YUV(int w, int h, unsigned char* bmp, unsigned int* yuv)
{
	unsigned int* u, * v, * y, * uu, * vv;
	unsigned int* pu1, * pu2, * pu3, * pu4;
	unsigned int* pv1, * pv2, * pv3, * pv4;
	unsigned char* r, * g, * b;
	int i, j;

	uu = new unsigned int[w * h];
	vv = new unsigned int[w * h];

	if (uu == NULL || vv == NULL)
		return 0;

	y = yuv;
	u = uu;
	v = vv;

	// Get r,g,b pointers from bmp image data....
	r = bmp;
	g = bmp + 1;
	b = bmp + 2;


	//Get YUV values for rgb values...

	for (i = 0; i < h; i++)
		for (j = 0; j < w; j++)
		{
			*y++ = (RGB2YUV_YR[*r] + RGB2YUV_YG[*g] + RGB2YUV_YB[*b] + 1048576) >> 16;
			*u++ = (-RGB2YUV_UR[*r] - RGB2YUV_UG[*g] + RGB2YUV_UBVR[*b] + 8388608) >> 16;
			*v++ = (RGB2YUV_UBVR[*r] - RGB2YUV_VG[*g] - RGB2YUV_VB[*b] + 8388608) >> 16;

			r += 3;
			g += 3;
			b += 3;
		}



	// Now sample the U & V to obtain YUV 4:2:0 format

	// Sampling mechanism...
	/*    @  ->  Y
		  #  ->  U or V
		  
		  @   @   @   @
			#   	#
		  @   @   @   @
		
		  @   @   @   @
			#   	#
		  @   @   @   @
	*/

	// Get the right pointers...
	u = yuv + w * h;
	v = u + (w * h) / 4;

	// For U
	pu1 = uu;
	pu2 = pu1 + 1;
	pu3 = pu1 + w;
	pu4 = pu3 + 1;

	// For V
	pv1 = vv;
	pv2 = pv1 + 1;
	pv3 = pv1 + w;
	pv4 = pv3 + 1;

	// Do sampling....
	for (i = 0; i < h; i += 2)
	{
		for (j = 0; j < w; j += 2)
		{
			*u++ = (*pu1 + *pu2 + *pu3 + *pu4) >> 2;
			*v++ = (*pv1 + *pv2 + *pv3 + *pv4) >> 2;

			pu1 += 2;
			pu2 += 2;
			pu3 += 2;
			pu4 += 2;

			pv1 += 2;
			pv2 += 2;
			pv3 += 2;
			pv4 += 2;
		}

		pu1 += w;
		pu2 += w;
		pu3 += w;
		pu4 += w;

		pv1 += w;
		pv2 += w;
		pv3 += w;
		pv4 += w;
	}


	delete uu;
	delete vv;

	return 1;
}







//
//Initialize conversion table for YUV420 to RGB
//
void InitConvertTable()
{
	long int crv, cbu, cgu, cgv;
	int i, ind;   

	crv = 104597; cbu = 132201;  /* fra matrise i global.h */
	cgu = 25675;  cgv = 53279;

	for (i = 0; i < 256; i++)
	{
		crv_tab[i] = (i - 128) * crv >> 16;
		cbu_tab[i] = (i - 128) * cbu >> 16;
		cgu_tab[i] = (i - 128) * cgu >> 16;
		cgv_tab[i] = (i - 128) * cgv >> 16;
		tab_76309[i] = 384 + ((76309 * (i - 16)) >> 16);
	}

	for (i = 0; i < 512 && i < 1024; i++)
		clp[i] = 0;
	for (i = 512 ; i < 1024; i++)
		clp[i] = 255;
	ind = 385;
	for (i = 0; i < 256; i++)
		clp[ind++] = i;
	for (i = 0; i < 1024; i++)
		if(clp[i]<0)
			clp[i] = 0;
	for (i = 0; i < 1024; i++)
		if(clp[i]>240)
			clp[i] = 240;
}

//
//  Convert from YUV420 to YUV422
//
void ConvertYUV4202YUV422(unsigned char* srcU, unsigned char* srcV, unsigned char* dstU, unsigned char* dstV, int width, int height)
{
	int i, j;
	unsigned char* puin1, * puin2, * puin3;
	unsigned char* pvin1, * pvin2, * pvin3;
	unsigned char* puout1, * puout2;
	unsigned char* pvout1, * pvout2;

	//Interligne vertical

	puin1 = srcU;
	puin2 = puin1 + 1;
	puin3 = puin2 + 1;

	pvin1 = srcV;
	pvin2 = pvin1 + 1;
	pvin3 = pvin2 + 1;

	puout1 = dstU;
	puout2 = puout1 + 1;

	pvout1 = dstV;
	pvout2 = pvout1 + 1;

	for (j = 0; j < height; j += 2)
	{
		for (i = 0; i < width; i += 2)
		{
			*puout1 = *puin1;
			*pvout1 = *pvin1;

			*puout2 = Limite8((9 * (*puin1 + *puin2) - (*puin1+*puin3) + 8)>>4);
			*pvout2 = Limite8((9 * (*pvin1 + *pvin2) - (*pvin1+*pvin3) + 8)>>4);

			puin1++;
			puin2++;

			pvin1++;
			pvin2++;

			puin3++;
			pvin3++;

			puout1+=2;
			puout2+=2;

			pvout1+=2;
			pvout2+=2;
/*
			if(i==width-6) {
				puin1-=(2 * width);
				puin2-=(4 * width);

				pvin1-=(2 * width);
				pvin2-=(4 * width);
			}
			if(i==width-4) {
				puin1++;
				puin2=puin1;

				pvin1-=2;
				pvin2=pvin1;
			}*/
		}
	}
}

//
//  Convert from YUV422 to YUV444
//
void ConvertYUV4222YUV444(unsigned char* srcU, unsigned char* srcV, unsigned char* dstU, unsigned char* dstV, int width, int height)
{
	int i, j;
	unsigned char* puin1, * puin2, * puin3;
	unsigned char* pvin1, * pvin2, * pvin3;
	unsigned char* puout1, * puout2;
	unsigned char* pvout1, * pvout2;

	puin1 = srcU;
	puin2 = puin1 + width;
	puin3 = puin2 + width;

	pvin1 = srcV;
	pvin2 = pvin1 + width;
	pvin3 = pvin2 + width;

	puout1 = dstU;
	puout2 = puout1 + width;

	pvout1 = dstV;
	pvout2 = pvout1 + width;

	for (j = 0; j < height; j += 2)
	{
		for (i = 0; i < width; i ++)
		{
			*puout1 = *puin1;
			*pvout1 = *pvin1;

			*puout2 = Limite8((9 * (*puin1 + *puin2) - (*puin1+*puin3) + 8)>>4);
			*pvout2 = Limite8((9 * (*pvin1 + *pvin2) - (*pvin1+*pvin3) + 8)>>4);

			puin1++;
			puin2++;

			pvin1++;
			pvin2++;

			puin3++;
			pvin3++;

			puout1++;
			puout2++;

			pvout1++;
			pvout2++;
		}
		if(j==height-6) {
			puin1-=(2 * width);
			puin2-=(4 * width);

			pvin1-=(2 * width);
			pvin2-=(4 * width);
		}
		if(j==height-4) {
			puin1-=(2 * width);
			puin2=puin1;

			pvin1-=(2 * width);
			pvin2=pvin1;
		}
		puout1 += width;
		puout2 += width;

		pvout1 += width;
		pvout2 += width;
	}
}

void DoubleYUV444(unsigned char* srcY, unsigned char* srcU, unsigned char* srcV, unsigned char* dstY, unsigned char* dstU, unsigned char* dstV, int width, int height)
{
	int i, j;
	unsigned char* pyin1, * pyin2, * pyin3;
	unsigned char* puin1, * puin2, * puin3;
	unsigned char* pvin1, * pvin2, * pvin3;
	unsigned char* pyout1, * pyout2;
	unsigned char* puout1, * puout2;
	unsigned char* pvout1, * pvout2;
	unsigned char* yy, * uu, * vv;

	yy = new unsigned char[4 * width * height];
	uu = new unsigned char[4 * width * height];
	vv = new unsigned char[4 * width * height];

	//Interligne vertical

	pyin1 = srcY;
	pyin2 = pyin1 + 1;
	pyin3 = pyin2 + 1;

	puin1 = srcU;
	puin2 = puin1 + 1;
	puin3 = puin2 + 1;

	pvin1 = srcV;
	pvin2 = pvin1 + 1;
	pvin3 = pvin2 + 1;

	pyout1 = yy;
	pyout2 = pyout1 + 1;

	puout1 = uu;
	puout2 = puout1 + 1;

	pvout1 = vv;
	pvout2 = pvout1 + 1;

	for (j = 0; j < height; j ++)
	{
		for (i = 0; i < width; i ++)
		{
			*pyout1 = *pyin1;
			*puout1 = *puin1;
			*pvout1 = *pvin1;

			*pyout2 = Limite8((9 * (*pyin1 + *pyin2) - (*pyin1+*pyin3) + 8)>>4);
			*puout2 = Limite8((9 * (*puin1 + *puin2) - (*puin1+*puin3) + 8)>>4);
			*pvout2 = Limite8((9 * (*pvin1 + *pvin2) - (*pvin1+*pvin3) + 8)>>4);

			pyin1++;
			pyin2++;
			pyin3++;

			puin1++;
			puin2++;
			puin3++;

			pvin1++;
			pvin2++;
			pvin3++;

			pyout1+=2;
			pyout2+=2;

			puout1+=2;
			puout2+=2;

			pvout1+=2;
			pvout2+=2;

			/*if(i==width-6) {
				puin1-=(2 * width);
				puin2-=(4 * width);

				pvin1-=(2 * width);
				pvin2-=(4 * width);
			}
			if(i==width-4) {
				puin1++;
				puin2=puin1;

				pvin1-=2;
				pvin2=pvin1;
			}*/
		}
	}

	pyin1 = yy;
	pyin2 = pyin1 + 2 * width;
	pyin3 = pyin2 + 2 * width;

	puin1 = uu;
	puin2 = puin1 + 2 * width;
	puin3 = puin2 + 2 * width;

	pvin1 = vv;
	pvin2 = pvin1 + 2 * width;
	pvin3 = pvin2 + 2 * width;

	pyout1 = dstY;
	pyout2 = pyout1 + 2 * width;

	puout1 = dstU;
	puout2 = puout1 + 2 * width;

	pvout1 = dstV;
	pvout2 = pvout1 + 2 * width;

	for (j = 0; j < height; j ++)
	{
		for (i = 0; i < 2 * width; i ++)
		{
			*pyout1 = *pyin1;
			*puout1 = *puin1;
			*pvout1 = *pvin1;

			*pyout2 = Limite8((9 * (*pyin1 + *pyin2) - (*pyin1+*pyin3) + 8)>>4);
			*puout2 = Limite8((9 * (*puin1 + *puin2) - (*puin1+*puin3) + 8)>>4);
			*pvout2 = Limite8((9 * (*pvin1 + *pvin2) - (*pvin1+*pvin3) + 8)>>4);
			
			pyin1++;
			pyin2++;

			puin1++;
			puin2++;

			pvin1++;
			pvin2++;

			puin3++;
			pvin3++;

			pyout1++;
			pyout2++;

			puout1++;
			puout2++;

			pvout1++;
			pvout2++;
		}
		if(j==height-6) {
			pyin1-=(2 * width);
			pyin2-=(4 * width);

			puin1-=(2 * width);
			puin2-=(4 * width);

			pvin1-=(2 * width);
			pvin2-=(4 * width);
		}
		if(j==height-4) {
			pyin1-=(2 * width);
			pyin2=pyin1;

			puin1-=(2 * width);
			puin2=puin1;

			pvin1-=(2 * width);
			pvin2=pvin1;
		}
		pyout1 += (2 * width);
		pyout2 += (2 * width);

		puout1 += (2 * width);
		puout2 += (2 * width);

		pvout1 += (2 * width);
		pvout2 += (2 * width);
	}
	delete yy;
	delete uu;
	delete vv;
}
//
//  Convert from YUV444 to RGB24
//
void ConvertYUV4442RGB(unsigned char* srcY, unsigned char* srcU, unsigned char* srcV, unsigned char* dst_ori, int width, int height)
{
	int i, j;

	//Rendu affichage
	int y1, u, v; 
	unsigned char* py1, * py2, * puu, * pvv;
	int c1, c2, c3, c4;
	unsigned char* d1, * d2;

	puu=srcU;
	pvv=srcV;

	py1 = srcY;
	py2 = py1 + width;

	d1 = dst_ori + 3 * width * height - 3 * width;
	d2 = d1 - 3 * width;
	for (j = 0; j < height; j ++)
	{
		for (i = 0; i < width; i ++)
		{
			u = *puu++;
			v = *pvv++;

			c4 = crv_tab[v];
			c2 = cgu_tab[u];
			c3 = cgv_tab[v];
			c1 = cbu_tab[u];

			//up-left
			y1 = tab_76309[*py1++]; 
			*d1++ = clp[Limite1024(y1 + c1)];  
			*d1++ = clp[Limite1024(y1 - c2 - c3)];
			*d1++ = clp[Limite1024(y1 + c4 )];
		}
		d1 -= 6 * width;
		d2 -= 6 * width;
	}
}

